/**************************************************************************************************/
/**************************************************************************************************/

#include "glut.h"
#include "subdivSurface.h"

/**************************************************************************************************/
/**************************************************************************************************/

int resX,resY;
int mouseX,mouseY,oldMouseX,oldMouseY,mouseButton;
int angleX,angleY,trans;

subdivSurface * surf;

bool move;

/**************************************************************************************************/
/**************************************************************************************************/

void drawReference() {

  glColor3f(0.7,0.7,0.7);
  for(int i=0;i<=100;i+=10) {
      glBegin(GL_LINES);
      glVertex3f(i,0,0);
      glVertex3f(i,100,0);
      glVertex3f(0,i,0);
      glVertex3f(100,i,0);
      glEnd();
    }

  glColor3f(1.0,0.0,0.0);
  glBegin(GL_LINES);
  glVertex3f(-10,0,0);
  glVertex3f(110,0,0);
  glEnd();

  glColor3f(0.0,1.0,0.0);
  glBegin(GL_LINES);
  glVertex3f(0,-10,0);
  glVertex3f(0,110,0);
  glEnd();

  glColor3f(0.0,0.0,1.0);
  glBegin(GL_LINES);
  glVertex3f(0,0,-10);
  glVertex3f(0,0,110);
  glEnd();

}

/**************************************************************************************************/
/**************************************************************************************************/

void setProjection() {

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective(60,(float)resX/(float)resY,1,256);

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  gluLookAt(0.0,50.0,100.0,0.0,0.0,0.0,0.0,0.0,1.0);	
  glTranslated(0, trans*0.25, trans);
  glRotated( angleX, 0,0,-1 ); 
  glRotated( angleY, cos(angleX*M_PI/180.0),sin(angleX*M_PI/180.0),0); 
  glTranslated(-50,-50,0);

}

/**************************************************************************************************/
/**************************************************************************************************/

void Init() {

    static float ambient[] = {0.1, 0.1, 0.1, 1.0};
    static float diffuse[] = {0.5, 1.0, 1.0, 1.0};
    static float position0[] = {0.0,0.0, 20.0, 0.0};
    static float position1[] = {0.0, 0.0, -20.0, 0.0};
    static float front_mat_shininess[] = {60.0};
    static float front_mat_specular[] = {0.2, 0.2, 0.2, 1.0};
    static float front_mat_diffuse[] = {0.5, 0.28, 0.38, 1.0};
    static float lmodel_ambient[] = {1.0, 1.0, 1.0, 1.0};
    static float lmodel_twoside[] = {GL_TRUE};

    glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
    glLightfv(GL_LIGHT0, GL_POSITION, position0);
    glEnable(GL_LIGHT0);
    
    glLightfv(GL_LIGHT1, GL_AMBIENT, ambient);
    glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuse);
    glLightfv(GL_LIGHT1, GL_POSITION, position1);
    glEnable(GL_LIGHT1);
    
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
    glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE, lmodel_twoside);

    glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, front_mat_shininess);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, front_mat_specular);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, front_mat_diffuse);

    glClearColor (0., 0., 0., 1.);
    glEnable(GL_DEPTH_TEST); 
    glDisable(GL_SMOOTH);

}

/**************************************************************************************************/
/**************************************************************************************************/

static void draw( void ) {

  int num;
  
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  setProjection();

  glDisable(GL_LIGHTING);
  drawReference();

  glEnable(GL_LIGHTING);
  surf->drawSurface();

  glutSwapBuffers();

}

/**************************************************************************************************/
/**************************************************************************************************/

static void mouse(int button, int state, int x, int y) {

  mouseX = x;
  mouseY = y;
	
  if (button==GLUT_LEFT_BUTTON && state == GLUT_DOWN) 
    mouseButton = 1;		
  else if  (button==GLUT_RIGHT_BUTTON && state == GLUT_DOWN)
    mouseButton = 3;
  else if  (button==GLUT_MIDDLE_BUTTON && state == GLUT_DOWN)
    mouseButton = 2;
  
  if  (state == GLUT_UP)
    mouseButton = 0;
  
  glutPostRedisplay();  

}

/**************************************************************************************************/
/**************************************************************************************************/

static void motion( int x, int y ) {

  oldMouseX = mouseX;
  oldMouseY = mouseY;
	
  mouseX = x;
  mouseY = y;

  if (!move) {
    if (mouseButton==1) {
      angleX += oldMouseX-mouseX;
      angleY += oldMouseY-mouseY;
    }
    
    if (mouseButton==3)
      trans += oldMouseX-mouseX+oldMouseY-mouseY;
  }
  else {
    if (mouseButton==1) {
      surf->moveSelectedPointX(mouseX-oldMouseX);
      surf->moveSelectedPointY(oldMouseY-mouseY);
    }
    else if (mouseButton==3)
      surf->moveSelectedPointZ(oldMouseY-mouseY);
  }

  glutPostRedisplay();
  
}

/**************************************************************************************************/
/**************************************************************************************************/

static void key( unsigned char key, int x, int y ) {

   switch (key) {
   case 27:               /* sortie si ESC */
     exit(0);
   case '+' :    
     surf->subdivideOneTime();
     break;
   case '-' : 
     surf->undoLastSubdivision();
     break;
   case 's' :
     surf->setSmooth(true);
     break;
   case 'f' :
     surf->setSmooth(false);
	 break;
   case ' ':
     move = !move;
     break;
   case 'h':
     surf->setShowControlPoints(false);
     break;
   case 'c':
     surf->setShowControlPoints(true);
     break;
   }

   glutPostRedisplay();

}

/**************************************************************************************************/
/**************************************************************************************************/

static void specialKey( int key, int x, int y )
{

   switch (key) {
   case GLUT_KEY_LEFT:
     surf->nextColumn();
     break;
   case GLUT_KEY_RIGHT:
     surf->previousColumn();
     break;
   case GLUT_KEY_UP:
     surf->nextRow();
     break;
   case GLUT_KEY_DOWN:
     surf->previousRow();
     break;
   default:
      return;
   }
 
   glutPostRedisplay();
}

/**************************************************************************************************/
/**************************************************************************************************/

static void reshape( int width, int height ) {

  resX = width;
  resY = height;

  glViewport(0, 0, width, height);

  setProjection();

  glutPostRedisplay();

}

/**************************************************************************************************/
/**************************************************************************************************/

int main( int argc, char *argv[]) {
   
	cout << "##################################" << endl;
	cout << "#  Subdivisions surfaces         #" << endl;
	cout << "##################################" << endl << endl;

	cout << " KEY : " << endl;
	cout << "+ : Sudvivide" << endl;
	cout << "- : Undo last Sudvivision" << endl;
	cout << "s : Smooth shading" <<endl;
	cout << "f : Flat shading" <<endl;		
	cout << "c : Swap control point displaying" <<endl;   
	cout << "space : Swap between control point and camera moving" <<endl;
	cout << "Arrow key : Move selected control point" <<endl << endl;   

  resX = 1024;
  resY = 768;
 
  mouseX      = 0;
  mouseY      = 0;
  oldMouseX   = 0;
  oldMouseY   = 0;
  mouseButton = 0;
  angleX      = 0;
  angleY      = 0;
  trans       = 0;

  move = false;

  surf = new subdivSurface(10,10);

  glutInitWindowPosition(100, 100);
  glutInitWindowSize(resX,resY);
  glutInitDisplayMode( GLUT_RGB|GLUT_DOUBLE );
  
  if (glutCreateWindow("Surfaces de subdivision") <= 0) {
    exit(0);
  }
  
  Init();

  glutReshapeFunc( reshape );
  glutKeyboardFunc( key );
  glutSpecialFunc( specialKey );
  glutDisplayFunc( draw );
  glutMouseFunc( mouse );
  glutMotionFunc( motion );
  glutMainLoop();
  
  return 0;

}
  

/**************************************************************************************************/
/**************************************************************************************************/
