
#include "subdivSurface.h"
#include "glut.h"

/**************************************************************************************************/
/**************************************************************************************************/

subdivSurface::subdivSurface(int patchLength, int patchWidth) { 

  patchW = patchWidth;
  patchL = patchLength;

  pointW = patchW;
  pointL = patchL;

  nbSubdivisions  = 0;
  nbControlPoints = 0;

  selectedPoint = 0;

  patchOk            = false;
  smooth             = false;
  showControlPoints  = false;

  controlPoints      = new struct coord3D[patchW*patchL];
  interpolatedPoints = new struct coord3D[patchW*patchL];
  tmpPoints          = NULL;

  for(int i=0;i<patchL;i++)
    for(int j=0;j<patchW;j++)
      addControlPoint(i*10,j*10,sin(i+j*j)*cos(j+i)*20+20);

  memcpy(interpolatedPoints,controlPoints,patchW*patchL*sizeof(struct coord3D));
  
  //cout << "Nombres de points dans le maillage : " << pointW*pointL << endl;

}

/**************************************************************************************************/
/**************************************************************************************************/

subdivSurface::~subdivSurface() {

  if (controlPoints != NULL)
    delete controlPoints;

  if (interpolatedPoints != NULL)
    delete interpolatedPoints;

  if (tmpPoints != NULL)
    delete tmpPoints;
	
}

/**************************************************************************************************/
/**************************************************************************************************/

void subdivSurface::setSmooth(bool s) {

  smooth = s;

  if (smooth)
    glEnable(GL_SMOOTH);
  else
    glDisable(GL_SMOOTH);

}

/**************************************************************************************************/
/**************************************************************************************************/

void subdivSurface::setShowControlPoints(bool s) {

  showControlPoints = s;

}

/**************************************************************************************************/
/**************************************************************************************************/

bool subdivSurface::addControlPoint(double x,double y,double z) {

  if (nbControlPoints<patchW*patchL) {
    controlPoints[nbControlPoints].x = x;
    controlPoints[nbControlPoints].y = y;
    controlPoints[nbControlPoints].z = z;
    nbControlPoints++;
    if (nbControlPoints==patchW*patchL)
      patchOk = true;
    return true;
  }
  else
    return false;
	
}

/**************************************************************************************************/
/**************************************************************************************************/

void subdivSurface::nextColumn() {

    if (selectedPoint>patchL)
      selectedPoint-=patchL;

}

/**************************************************************************************************/
/**************************************************************************************************/

void subdivSurface::previousColumn() {

  if (selectedPoint<patchW*patchL-patchL)
    selectedPoint+=patchL;

}

/**************************************************************************************************/
/**************************************************************************************************/

void subdivSurface::nextRow() {

  if (selectedPoint<patchW*patchL-1)
    selectedPoint++;

}

/**************************************************************************************************/
/**************************************************************************************************/

void subdivSurface::previousRow() {

  if (selectedPoint>0)
    selectedPoint--;

}

/**************************************************************************************************/
/**************************************************************************************************/

void subdivSurface::moveSelectedPointX(int x) {

  controlPoints[selectedPoint].x += x;

  reSubdivide(nbSubdivisions);

}

/**************************************************************************************************/
/**************************************************************************************************/

void subdivSurface::moveSelectedPointY(int y) {
 
  controlPoints[selectedPoint].y += y;

  reSubdivide(nbSubdivisions);

}

/**************************************************************************************************/
/**************************************************************************************************/

void subdivSurface::moveSelectedPointZ(int z) {

  controlPoints[selectedPoint].z += z;

  reSubdivide(nbSubdivisions);

}

/**************************************************************************************************/
/**************************************************************************************************/

bool subdivSurface::subdivideOneTime() {

  int n;
  int offset;

  if (tmpPoints==NULL)
    delete tmpPoints;

  tmpPoints = new struct coord3D[pointW*(pointL-1)*2];

  n = 0;
  for(int j=0;j<pointW;j++)
    for(int i=0;i<pointL-1;i++) {
      offset = j*pointL+i;
      tmpPoints[n].x = 0.75*interpolatedPoints[offset].x+0.25*interpolatedPoints[offset+1].x;
      tmpPoints[n].y = 0.75*interpolatedPoints[offset].y+0.25*interpolatedPoints[offset+1].y;
      tmpPoints[n].z = 0.75*interpolatedPoints[offset].z+0.25*interpolatedPoints[offset+1].z;
      tmpPoints[n+1].x = 0.25*interpolatedPoints[offset].x+0.75*interpolatedPoints[offset+1].x;
      tmpPoints[n+1].y = 0.25*interpolatedPoints[offset].y+0.75*interpolatedPoints[offset+1].y;
      tmpPoints[n+1].z = 0.25*interpolatedPoints[offset].z+0.75*interpolatedPoints[offset+1].z;
      n+=2;
    }

  if (interpolatedPoints==NULL)
    delete interpolatedPoints;

  pointL = (pointL-1)*2;

  interpolatedPoints = new struct coord3D[(pointW-1)*2*pointL];

  for(int i=0;i<pointL;i++) {
    n = i;
    for(int j=0;j<pointW-1;j++) {
      offset = j*pointL+i;
      interpolatedPoints[n].x = 0.75*tmpPoints[offset].x+0.25*tmpPoints[offset+pointL].x;
      interpolatedPoints[n].y = 0.75*tmpPoints[offset].y+0.25*tmpPoints[offset+pointL].y;
      interpolatedPoints[n].z = 0.75*tmpPoints[offset].z+0.25*tmpPoints[offset+pointL].z;
      interpolatedPoints[n+pointL].x = 0.25*tmpPoints[offset].x+0.75*tmpPoints[offset+pointL].x;
      interpolatedPoints[n+pointL].y = 0.25*tmpPoints[offset].y+0.75*tmpPoints[offset+pointL].y;
      interpolatedPoints[n+pointL].z = 0.25*tmpPoints[offset].z+0.75*tmpPoints[offset+pointL].z;
      n+=2*pointL;
    }
  }

  pointW = (pointW-1)*2;

  nbSubdivisions++;

 // cout << "Nombres de points dans le maillage : " << pointW*pointL << endl;

  return true;

}

/**************************************************************************************************/
/**************************************************************************************************/

bool subdivSurface::undoLastSubdivision() {

  reSubdivide(nbSubdivisions-1);

  return true;

}

/**************************************************************************************************/
/**************************************************************************************************/

bool subdivSurface::reSubdivide(int n) {

  if (interpolatedPoints==NULL)
    delete interpolatedPoints;

  interpolatedPoints = new struct coord3D[patchW*patchL];
  pointW = patchW;
  pointL = patchL;

  memcpy(interpolatedPoints,controlPoints,patchW*patchL*sizeof(struct coord3D));

  nbSubdivisions = 0;
 
  for(int i=0;i<n;i++)
    subdivideOneTime();

  return true;

}

/**************************************************************************************************/
/**************************************************************************************************/

void subdivSurface::getPoint(int x,int y,struct coord3D * point) {

  point->x = 0.0;
  point->y = 0.0;
  point->z = 0.0;

  if (x>=0 && x<pointL && y>=0 && y<pointW) { 
    point->x = interpolatedPoints[y*pointL+x].x;
    point->y = interpolatedPoints[y*pointL+x].y;
    point->z = interpolatedPoints[y*pointL+x].z;
  }

}

/**************************************************************************************************/
/**************************************************************************************************/

void subdivSurface::getFaceNormal(int x,int y,struct coord3D * normal) {

  struct coord3D v1;
  struct coord3D v2;
  struct coord3D p1;
  struct coord3D p2;
  double l;

  normal->x = 0;
  normal->y = 0;
  normal->z = 0;

  if (x>=0 && x<pointL-1 && y>=0 && y<pointW-1) {
    getPoint(x,y,&p1);
    getPoint(x+1,y,&p2);
    v1.x = p2.x-p1.x; 
    v1.y = p2.y-p1.y;
    v1.z = p2.z-p1.z;
    getPoint(x,y+1,&p2);
    v2.x = p2.x-p1.x; 
    v2.y = p2.y-p1.y;
    v2.z = p2.z-p1.z;
    normal->x = v1.y*v2.z-v2.y*v1.z;
    normal->y = v2.x*v1.z-v1.x*v2.z;
    normal->z = v1.x*v2.y-v2.x*v1.y;
    
    l = sqrt(normal->x*normal->x+normal->y*normal->y+normal->z*normal->z);

    normal->x /= l;
    normal->y /= l;
    normal->z /= l;

  }

}

/**************************************************************************************************/
/**************************************************************************************************/

void subdivSurface::getVertexNormal(int x,int y,struct coord3D * normal) {
  
  struct coord3D n1;
  struct coord3D n2;
  struct coord3D n3;
  struct coord3D n4;
  double l;

  getFaceNormal(x,y,&n1);
  getFaceNormal(x-1,y,&n2);
  getFaceNormal(x-1,y-1,&n3);
  getFaceNormal(x,y-1,&n4);

  normal->x = (n1.x + n2.x + n3.x + n4.x)/4.0;
  normal->y = (n1.y + n2.y + n3.y + n4.y)/4.0;
  normal->z = (n1.z + n2.z + n3.z + n4.z)/4.0;

  l = sqrt(normal->x*normal->x+normal->y*normal->y+normal->z*normal->z);

  normal->x /= l;
  normal->y /= l;
  normal->z /= l;

}

/**************************************************************************************************/
/**************************************************************************************************/

void subdivSurface::drawSurface() {

  struct coord3D p1;
  struct coord3D p2;
  struct coord3D p3;
  struct coord3D p4;

  struct coord3D n1;
  struct coord3D n2;
  struct coord3D n3;
  struct coord3D n4;

  if (patchOk) {

    glEnable(GL_LIGHTING);

    glBegin(GL_QUADS);

    for(int i=0;i<pointL-1;i++)
      for(int j=0;j<pointW-1;j++) {

	getPoint(i,j,&p1);
	getPoint(i+1,j,&p2);
	getPoint(i+1,j+1,&p3);
	getPoint(i,j+1,&p4);

	if (smooth) {
	  getVertexNormal(i,j,&n1);
	  getVertexNormal(i+1,j,&n2);
	  getVertexNormal(i+1,j+1,&n3);
	  getVertexNormal(i,j+1,&n4);

	  glNormal3f(n1.x,n1.y,n1.z);
	  glVertex3f(p1.x,p1.y,p1.z);

	  glNormal3f(n2.x,n2.y,n2.z);
	  glVertex3f(p2.x,p2.y,p2.z);

	  glNormal3f(n3.x,n3.y,n3.z);
	  glVertex3f(p3.x,p3.y,p3.z);

	  glNormal3f(n4.x,n4.y,n4.z);
	  glVertex3f(p4.x,p4.y,p4.z);
	}
	else {
	  getFaceNormal(i,j,&n1);
	  glNormal3f(n1.x,n1.y,n1.z);
	  glVertex3f(p1.x,p1.y,p1.z);
	  glVertex3f(p2.x,p2.y,p2.z);
	  glVertex3f(p3.x,p3.y,p3.z);
	  glVertex3f(p4.x,p4.y,p4.z);
	}

      }

    glEnd();

    if (showControlPoints) {
      glDisable(GL_LIGHTING); 
      glPushMatrix();
      glColor3f(1.0,1.0,0.0);
      glTranslated(controlPoints[selectedPoint].x,controlPoints[selectedPoint].y,controlPoints[selectedPoint].z);
      glutSolidCube(2); 
      glPopMatrix();
      glColor3f(1.0,0.0,0.0);
      for(int i=0;i<patchL*patchW;i++) {
	glPushMatrix();
	glTranslated(controlPoints[i].x,controlPoints[i].y,controlPoints[i].z);
	glutSolidCube(1);
	glPopMatrix();
      }
    }

  }

}

/**************************************************************************************************/
/**************************************************************************************************/
